/*
 *  vines.c
 *
 *  2.0 Revamped, added to, ported, optimized, etc. by Miles O'Neal
 *
 *  1.1 Updated to work on X11 March 1990 by Dan Heller.
 *
 *  1.0 Written by Dan Heller one fine winter week in the spirit of the
 *  coming holidays, 1985.  Works on Sun versions 1.4 and up.
 *
 *  The differences between X11 and Sunview are handled by whether the
 *  compile-time switch X11 is #defined.  In some cases different code
 *  is used, or code is added or deleted based on this switch.  In a few
 *  cases, certain X calls are just placed where they should be for X,
 *  but with SunView they are #defined to not exist.  There are a couple
 *  of cases where a switch is tested that can only be set or reset under
 *  one graphics system, but the default works for the other, so it is
 *  not wrapped in extraneous #if tests. -Miles
 *
 *
 * ORIGINAL NOTES (somewhat SunView centric)
 *
 * The way vines works is a "stem" grows and leaves grow from that stem.
 * The stem and the leaf are arrays of "points" (struct pr_pos).  You can
 * add your own images (like flowers or something) with the -S and -L
 * flags.  The usages for these flags are:
 * 
 * -S file
 * -L file
 * 
 * These files are opened and pairs of points are read in place of the
 * Stems and Leaves respectively.  The end of an image is represented
 * by the coordinate pair 0,0.
 * 
 * For each coordinate pair read, pr_vectors are drawn from 0,0 to the
 * first point, then to the second and so on till the final 0,0 and then
 * a pr_polygon2 routine is called to fill in the image.
 *
 * Because there is an extra line in the middle of leaves, some of the code
 * around pr_polygon2() is added to continue drawing the leaves.  Unless your
 * own images are simple, you're probably going to have to do the same thing
 * in that same area (search for pr_polygon2 for where I'm talking about).
 *
 * If you do anything interesting to this code, I would GREATLY appreciate
 * knowing what you did.  Please send any hacks or comments/flames to:
 *
 *    argv@sun.com
 */

char *err_msg[] = {
    "Usage: %s [-c [-H] | -v N | -e] [-C [-D] | -f [-D] | -r | -t N] \\\n",
    "     [-G N] [-b N] [-i N] [-l N] [-s N] [-R N] [-L file] [-S file] \\",
#ifdef X11
    "     [-F] [-x] [-d D]",
#else /* X11 */
    "     [-w] [-d D]",
#endif /* X11 */
    "     [-h | -? | -z]",
    "     [-g N]",
    "",
    "   -c:    vines grow from the center of the screen",
    "   -e:    vines grow along top & side edges of the screen (default)",
    "   -E:    vines grow along all edges of the screen",
    "   -v N:  grow N vines from the top of the screen. (default=4)",
    "   -H:    vines halt upon RETURN (use with -c)",
    "",
    "   -r:    reverse video; leaves are inverted",
    "   -C:    color (spring season, green leaves)",
    "   -f:    fall season (brown leaves: color only)",
    "   -D:    degrades the color down the colormap (use with -C or -f)",
    "   -t N:  run in technicolor, with N colors",
    "   -P file:  use palette of colors found in file",
    "",
    "   -G N:  growth rate of 1 to 100, where 100 is like kudzu (default: 100)",
    "   -b N:  probability of branching (max 40)",
    "   -i N:  interleave 1 black leaf per N white leaves (-N inverts ratio)",
    "   -l N:  leaf density on stem (1 (least) to 35 (most))",
    "   -s N:  max size of leaves (default: 2)",
    "   -R N:  leaf rotation increment (max 360)",
    "",
    "   -L file:  use leaf defined in file as x,y coordinate pairs",
    "   -S file:  use stem defined in file as x,y coordinate pairs",
    "",
#ifdef X11
    "   -F:    fork subprocess for better speed (may have problems)",
    "   -x:    run in XSynchronize mode",
    "   -d D:  use display D where D is any standard X DISPLAY format",
#else /* X11 */
    "   -w:    vines grow around each window on screen (not icons; no color)",
    "   -d D:  use alternate frame buffer D; default is /dev/fb",
#endif /* X11 */
    "",
    "   -z:    show version information",
    "   -h:    show this help/usage message",
    "   -?:    show this help/usage message",
    "   -g N:  debug at level N (see man page for details)",
    0
};

#define MAX_FORKS 6

#include <stdio.h>
#include <sys/file.h>
#include <ctype.h>
#include <signal.h>
#include <stdlib.h>
#include <time.h>
#include <sys/time.h>
#include <unistd.h>
#include "version.h"

int getopt(int argc, char * const argv[], const char *optstring);

#ifdef X11
#   include <X11/Xos.h>
#   include <X11/Xatom.h>
#   include <X11/Xlib.h>
#   include <X11/Xutil.h>
    typedef XPoint point_t;
    struct rect {
        int r_left, r_top, r_width, r_height;
    };
#else /* X11 */
    typedef struct pr_pos point_t;
#   include <suntool/tool_hs.h>
#   include <pixrect/pixrect_hs.h>
#   include <suntool/wmgr.h>
#   include <sunwindow/cms_mono.h>
#   define DefaultDepthOfScreen(s) s->pr_depth
#   define WidthOfScreen(s) s->pr_size.x
#   define HeightOfScreen(s) s->pr_size.y
#   define WUF_WMGR1 0x10
#   define XFlush(dpy) /* not used in Suntools */
#   define XCloseDisplay(dpy) /* not used in Suntools */
#endif /* X11 */

#if defined(linux)
#   define _USE_BSD
#endif
#include <sys/wait.h>
#include <sys/resource.h>
#include <math.h>

#define UP                      001  /* turn object up-side down */
#define MIRROR                  002  /* mirror the object */
#define FLIP                    004  /* rotate object 90 deg. */
#define FLOP                    010  /* rotate object -90 deg. */
#define DEG_TO_RADS             57.295779513

#ifdef X11
#   define SetBlack() intensity = (is_color || !reverse)? \
        BlackPixelOfScreen(screen) : WhitePixelOfScreen(screen); \
        XSetForeground(dpy, gc, intensity)
#   define SetColor(color) intensity = is_color? color : \
        reverse? WhitePixelOfScreen(screen) : BlackPixelOfScreen(screen); \
        XSetForeground(dpy, gc, intensity)
#   define draw(x1,y1,x2,y2) \
        XDrawLine(dpy,win,gc,(int)x1,(int)y1,(int)x2,(int)y2)
#else /* X11 */
#   define SetBlack() (intensity = (is_color || !reverse)? \
        BLACK : is_color? 255 : WHITE)
#   define SetColor(color)     if (is_color) intensity = color; else SetBlack()
#   define RAS_OP              PIX_COLOR(intensity)|PIX_SRC    
#   define draw(x1,y1,x2,y2)   pr_vector(screen,x1,y1,x2,y2,RAS_OP,1)
#endif /* X11 */

/*
 * each {x,y} is a relative point to 0,0. End is {0,0} pair
 */
point_t leaf[] = {
    {-4,4}, {-9,6}, {-13,9}, {-15,11}, {-16,14}, {-17,16}, {-17,20},
    {-16,24}, {-13,28}, {-11,30}, {-9,32}, {-4,36}, {-1,38}, {4,39}, {6,37},
    {4,37}, {3,34}, {3,31}, {4,29}, {6,25}, {9,20}, {10,16}, {10,12},
    {8,8}, {7,7}, {4,6}, {2,4}, {1,0}, {0,2}, {-1,4}, {-3,8},
    {-5,12}, {-6,16}, {-6,20}, {-5,24}, {-3,27}, {0,0}
};
int leafsize = sizeof (leaf) / sizeof (point_t);

point_t stem[50] = {
    {0,7}, {1,8}, {2,7}, {1,12}, {0,11}, {-1,12}, {-3,15}, {-3,19}, {-2,22},
    {-1,24}, {1,26}, {3,28}, {4,30}, {5,32}, {5,36}, {7,40}, {8,42}, {9,43},
    {8,49}, {5,51}, {2,54}, {0,60}, {0,0}
};
int stemsize = sizeof (stem) / sizeof (point_t);

point_t leaves[360][sizeof(leaf) / sizeof(point_t)];
struct rect win_rects[32];

int reverse = 0;      /*  -r:  reverse video */
int rot_increment=45; /*  -R:  rotation increment... up to 360 degrees */
int interleaf = 3;    /*  -i:  interleaf black and white leaves */
int lsize = 2;        /*  -s:  max magnification variable for leaves */
int leaf_freq = 30;   /*  -l:  frequency of leafs per stem */
int do_halt = 0;      /*  -H:  program halts on RETURN */
int branch_freq = 5;  /*  -b:  how often to branch off a new vine */
int n_vines = 4;      /*  -v:  4 is default */
int from_center = 0;  /*  -c:  start at the center of screen */
int is_color = 0;     /*  -C:  user is running from a color monitor */
int dbug = 0;         /*  -g:  debug mode, inspired by cc's -g flag for dbx */
int do_edges = 1;     /*  -e:  just traverse the edges of the screen */
int all_edges = 0;    /*  -E:  traverse all edges of the screen */
int do_windows = 0;   /*  -w:  grow around each open window in suntools */
int season = 0;       /*  -f:  fall season */
int fade = 0;         /*  -D:  fade the color down the colormap */
int xsyn = 0;         /*  -x:  run in XSynchronize mode */
int do_fork = 0;      /*  -F:  fork processes (X11) */
int grow_rate = 100;  /*  -G:  growth rate */
int gr_sleep = 0;
int technicolor = 0;  /*  -t:  technicolor (default 0, which is off) */

unsigned long colors[255];
int max_color = 1;    /* maximum number of colors available */
int win_no, my_win;   /* number of windows running / which win we are */
int total_forks;      /* the total number of forks we've done so far */

char *prog_name;

#ifdef X11
    unsigned long intensity;        /* shade of color to paint */
    Display *dpy; /* display */
    Screen *screen;
    Window win; /* root window */
    GC gc; /* points to default GC of screen */
    char *device; /* default screen [default = NULL (see XOpenDisplay below)] */
#else /* X11 */
    int intensity;        /* shade of color to paint */
    struct pixrect *screen;   /*  our screen's frame buffer */
    char *device = "/dev/fb"; /* default frame buffer */
#endif /* X11 */

char *leaf_file = NULL;      /* file for alternate leaf image */
char *stem_file = NULL;      /* file for alternate stem image */
char *palette_file = NULL;   /* file for color palette */
FILE *f;

#ifndef X11
#   include <pixrect/pr_planegroups.h>
#endif /* X11 */

int check_rects(register int x, register int y);
int get_windows();
int x_error(Display *display, XErrorEvent *error);

void dead_kids();
void do_leaf(int x, int y);
void do_stem(int *x, int *y);
void draw_leaf(int x, int y, long rotation, double magnification, long mirror);
void draw_stem(register int *x, register int *y, long dir);
void edge_vines(int x, int y, int which_way);
void init_display();
void init_srandom();
void initialize_color();
void parse_args(int argc, char **argv);
void random_vines(int xstart, int ystart);
void rotate_obj(point_t *object, point_t *newobject, int npts, double rotation);
void rotate_pt(point_t * oldpt, point_t *newpt, double rot);


int main(int argc, char **argv)
{
    register int count, x, y = 1;
    int c;

    prog_name = argv[0];

    if (argc > 1)
        parse_args(argc, argv);
    else
        do_edges = 1, interleaf = 3;

    if (leaf_file) {
        if (!(f = fopen(leaf_file,"r")))
            perror(leaf_file), exit(1);
        for (c = 0; c < leafsize; ++c)
            if (fscanf(f,"%hd,%hd", &leaf[c].x, &leaf[c].y) < 2)
                break;
        (void) fclose(f);
        if (leaf[c - 1].y || leaf[c - 1].x) {
            if (c == leafsize)
                fprintf(stderr, "%s: too many points (limit is %d)\n",
                        leaf_file, c);
            else
                fprintf(stderr, "%s: last coordinate pair must be 0,0\n",
                      leaf_file);
            exit(1);
        }
        leafsize = c - 1;
    }

    if (stem_file) {
        if (!(f = fopen(stem_file,"r")))
            perror(stem_file), exit(1);
        for (c = 0; c < stemsize; ++c)
            if (fscanf(f,"%hd,%hd", &stem[c].x, &stem[c].y) < 2)
                break;
        (void) fclose(f);
        if (stem[c - 1].y || stem[c - 1].x) {
            if (c == stemsize)
                fprintf(stderr, "%s: too many points (limit is %d)\n",
                        stem_file, c);
            else
                fprintf(stderr, "%s: last coordinate pair must be 0,0\n",
                        stem_file);
            exit(1);
        }
    }

    if (do_halt || (do_fork && from_center))
        /* don't count this fork ... */
        if ((! do_fork) || fork())
            getchar(), kill(0, 9);
    if (interleaf < 0)
        reverse = !reverse, interleaf *= -1;
    setbuf(stdout, NULL);

    printf("growing leaves ");
    for (x = 0, y = 0; x < 360; x += rot_increment) {
        if (dbug > 0)
            printf("Making leaf at %d rotation\n", rot_increment);
        rotate_obj(leaf, leaves[y++], leafsize, x/DEG_TO_RADS);
        if (!(y % 10))
            putchar('.');
    }
    puts(" done.");

    init_display();
#ifndef X11
    if (!is_color && do_windows) {
        if (!get_windows())
            puts("No windows!");

        for (my_win = 0; my_win < win_no-1 && fork(); my_win++)
            ;

        if (fork())
            edge_vines(win_rects[my_win].r_left + win_rects[my_win].r_width / 2,
                            win_rects[my_win].r_top, FLIP);
        else
            edge_vines(win_rects[my_win].r_left + win_rects[my_win].r_width / 2,
                            win_rects[my_win].r_top, FLOP);
        exit(0);
    }
#endif /* X11 */
    y = 1;
    if (do_edges) {
        if (do_fork) {
            if (!fork()) {
                init_display();
                edge_vines(WidthOfScreen(screen) / 2, 10, FLIP);
            } else
                edge_vines(WidthOfScreen(screen) / 2, 10, FLOP);
        } else {
            edge_vines(WidthOfScreen(screen) / 2, 10, FLIP);
            edge_vines(WidthOfScreen(screen) / 2, 10, FLOP);
        }
        XCloseDisplay(dpy);
        exit(0);
    }
#ifdef SIGCHLD
    (void) signal(SIGCHLD, dead_kids);
#endif /* SIGCHLD */
    for (count = 0, x = WidthOfScreen(screen) / (n_vines+1);
      count < n_vines;
      x += WidthOfScreen(screen) / (n_vines+1), count++)
    {
        if (dbug) printf("vine %d\n", count);
        if (from_center)
            random_vines(WidthOfScreen(screen)/2, HeightOfScreen(screen)/2);
        else
            random_vines(x, y);
    }
    return 0;
}



void init_srandom()
{
    struct timeval tv;

    gettimeofday(&tv, NULL);
    srandom((int) ((tv.tv_sec + tv.tv_usec) % getpid()));
}



#ifdef SIGCHLD
/* catch exited forked processes and decrement number of total_forks avail */
void dead_kids()
{
#if defined(VINES_1_0)
    union wait status;
#else
    int status;
#endif

    while (wait3(&status, WNOHANG, (struct rusage *) 0))
        total_forks--;
}
#endif /* SIGCHLD */



void edge_vines(int x, int y, int which_way)
{
    register int stop_l, stop_r, y_stop;

#ifndef X11
    if (do_windows) {
        stop_l = win_rects[my_win].r_left;
        stop_r = stop_l + win_rects[my_win].r_width;
        y_stop = y + win_rects[my_win].r_height;
    } else {
#endif /* X11 */
        stop_r = WidthOfScreen(screen) - 60;
        stop_l = 60;
        y_stop = HeightOfScreen(screen) - 60;
#ifndef X11
    }
#endif /* X11 */

    /* srandom must be here for recursive calls from new forked procs */
    init_srandom();

    leaf_freq = 34;  /* edge vines have constant leaf frequency */

    while(x > stop_l && x < stop_r)
        draw_stem(&x, &y, which_way | ((random() % 2) * UP));
    while(y < y_stop)
        draw_stem(&x, &y, (random() % 2) * MIRROR);
    /* draw along the bottom */
    if (all_edges)
        while(x < stop_r)
            draw_stem(&x, &y,
                (random() % 2) * UP | ((which_way == FLIP) ? FLOP : FLIP));
    XFlush(dpy);
}



void random_vines(int xstart, int ystart)
{
    if (do_fork) {
        /* parent returns */
        if(total_forks < MAX_FORKS && fork()) {
            total_forks++;
            return;
        }
        init_display();
        /* make the child think there are too many forks */
        total_forks = MAX_FORKS;
        while(ystart >= 0 && ystart < HeightOfScreen(screen) &&
              xstart >= 0 && xstart < WidthOfScreen(screen)) {
            do_stem(&xstart, &ystart);
            if (!(random() % (1 + (40 - branch_freq))))
                random_vines(xstart, ystart);
        }
        XFlush(dpy);
        XCloseDisplay(dpy);
        exit(0);
    } else {
        while(ystart >= 0 && ystart < HeightOfScreen(screen) &&
              xstart >= 0 && xstart < WidthOfScreen(screen)) {
            do_stem(&xstart, &ystart);
            if (!(random() % (1 + (40 - branch_freq))))
                random_vines(xstart, ystart);
        }
    }
}



void do_stem(int *x, int *y)
{
    long dir = 0;
    if (!(random() % (from_center ? 2 : 20)))
        dir |= UP;
    if (random() % 2)
        dir |= MIRROR;
    if (random() % 2)
        dir |= (random() % 2) ? FLIP : FLOP;
    draw_stem(x, y, dir);
}



void draw_stem(register int *x, register int *y, long dir)
{
    /* ox,oy are old x and y; nx,ny are new x and new y coords */

    register int tmp, ox = *x, oy = *y, nx, ny;
    for(tmp = 0; (stem[tmp].y || stem[tmp].x) && tmp < stemsize; tmp++) {
        if (dir & FLIP) {
            ny = *y + (stem[tmp].x * ((dir & UP) ? -1 : 1));
            nx = *x + (stem[tmp].y * ((dir & MIRROR) ? -1 : 1));
        } else if (dir & FLOP) {
            ny = *y + (stem[tmp].x * ((dir & UP) ? -1 : 1));
            nx = *x - (stem[tmp].y * ((dir & MIRROR) ? -1 : 1));
        } else {
            ny = *y + (stem[tmp].y * ((dir & UP) ? -1 : 1));
            nx = *x + (stem[tmp].x * ((dir & MIRROR) ? -1 : 1));
        }
        if (!do_windows || check_rects(nx, ny)) {
            SetBlack();
            draw(ox, oy, nx, ny);
            if(!(random() % (1 + (35 - leaf_freq))))
                do_leaf(nx, ny);
        }
        ox = nx, oy = ny;
    }
    *x = ox, *y = oy;
}


void do_leaf(int x, int y)
{
    int reverse_it = 0;
    if (interleaf && (reverse_it = !(random() % interleaf)))
        reverse = !reverse;
    /* start at x,y and rotate */
    draw_leaf(x, y, (long)(random() % (long) (360/rot_increment)),
                (double)(.6 + (random() % (10 * lsize)) / 20.0),  /* magnify */
                (long) (random() % 2 ? -1 : 1)); /* mirror image */
    if (interleaf && reverse_it)
        reverse = !reverse;
}



void draw_leaf(int x, int y, long rotation, double magnification, long mirror)
{
    static col;
    int tmp, ox, oy, nx, ny;
    int i;
#ifndef X11
    int npts[1];
#endif
    point_t leaf2[sizeof(leaf) / sizeof(point_t)];
    double foo, foo2;

    ox = x;
    oy = y;
    if (rotation == 0 && dbug > 2) {
        printf("x=%d y=%d leafsize=%d rotation=%g magnification=%g mirror=%g\n",
            x, y, leafsize, (double)rotation, magnification, (double)mirror);
        printf("rotation=[%d,%d]\n", leaves[rotation][1].x,
            leaves[rotation][1].y);
    }

    /* magnify, move leaf into position, and mirror, if necessary */
    foo = magnification * (double)mirror;
    for(tmp = 0; tmp < leafsize; tmp++) {
        foo2 = foo;

        /* work around weird Dell/gcc 2.8.x bug -meo */
#ifdef WEIRD_BUG
        if (rotation == 0) {
            foo2 *= -9.0;
            ny = y + (int)(magnification * 6.0);
        } else {
#endif /* WEIRD_BUG */
            foo2 *= (double)leaves[rotation][tmp].x;
            ny = y + (int)(magnification * (double)leaves[rotation][tmp].y);
#ifdef WEIRD_BUG
        }
#endif /* WEIRD_BUG */

        nx = x + (int) foo2;
        leaf2[tmp].y = oy, leaf2[tmp].x = ox;

        if (rotation == 0 && tmp <= 1 && dbug > 2) {
            printf("ox=%d oy=%d nx=%d ny=%d foo=%g foo2=%g\n",
                ox, oy, nx, ny, foo, foo2);
        }
        ox = nx, oy = ny;
    }
    /* paint leaf before drawing outline */
    reverse = !reverse;
    if (is_color)
        if (fade) {
            col += fade;
            if (!col || col == max_color-1)
                fade = -fade;
        } else {
            col = random() % max_color;
            /* col = 80 + random() % 165; */
        }
    SetColor(colors[col]); /* for black and white: intensity has no meaning */
    reverse = !reverse;
#ifdef X11
    XFillPolygon(dpy, win, gc, leaf2, 28, Nonconvex, CoordModeOrigin);
#else /* X11 */
    npts[0] = 28;   /* kludge for middle line of leaf! */
    pr_polygon_2(screen, 0, 0, 1, npts, leaf2, RAS_OP, (Pixrect *) 0,0,0);
#endif /* X11 */

    if (rotation == 0 && dbug > 2) {
        for (i = 0; i < leafsize; i++)
            printf("[%d,%d] ", leaf2[i].x, leaf2[i].y);
        printf("\n");
    }
    if (dbug > 1)
        usleep(1000000);

    SetBlack();
    /* now draw the outline of the leaf */
#ifndef X11
    for (tmp -= 2; tmp; tmp--)
        draw(leaf2[tmp].x, leaf2[tmp].y, leaf2[tmp-1].x, leaf2[tmp-1].y);
#else /* X11 */
    XDrawLines(dpy, win, gc, leaf2, (tmp - 1), CoordModeOrigin);
#endif /* X11 */
    if (grow_rate < 100)
        XFlush(dpy);
    usleep(gr_sleep);
}



/* check to see if we're inside the rect of another window.  The win_rects
 * array has the oldest children (windows sitting "under" other windows)
 * first in the array.  So, we only check windows higher than us in the array
 * since we know we overlay the other windows.
 */
int check_rects(register int x, register int y)
{
#ifndef X11
    register int win;
    static lastwin = -1;
    for (win = my_win+1; win < win_no; win++)
        if (x > win_rects[win].r_left &&
            x < win_rects[win].r_left + win_rects[win].r_width &&
            y > win_rects[win].r_top &&
            y < win_rects[win].r_top + win_rects[win].r_height) {
                if (dbug && win != lastwin)
                    printf("win %d ran into win %d at %d,%d\n",
                        my_win, (lastwin = win), x,y);
                return 0;
            }
#endif /* X11 */
    return 1;
}



/* foreach pt in object, rotate_pt() */
void rotate_obj(point_t *object, point_t *newobject, int npts, double rotation)
{
    int i;
    for(i = 0; i < npts; i++)
        rotate_pt(&object[i], &newobject[i], rotation);
}



/* matrix multiplication */
void rotate_pt(point_t * oldpt, point_t *newpt, double rot)
{
    /* temp storage -- might want to copy to itself */
    register double tx = (double)oldpt->x, ty = (double)oldpt->y, X,Y;
    X = tx * cos(rot) - ty * sin(rot);
    Y = tx * sin(rot) + ty * cos(rot);
    newpt->x = (int)X + ((X * 10.0 - (int)X * 10) >= 5); /* round up */
    newpt->y = (int)Y + ((Y * 10.0 - (int)Y * 10) >= 5); /* round up */
    if (dbug == 2)
        printf("%g,%g --> %g,%g --> %d,%d\n", tx, ty, X, Y, newpt->x, newpt->y);
}



void parse_args(int argc, char **argv)
{
    register int c;
    extern char *optarg;

    int errors = 0;

    while ((c = getopt(argc, argv, "DEefFHhrwcCxzb:d:t:i:g:G:l:s:v:R:L:P:S:?"))
      != EOF)
        switch (c) {
            case 'c':
                from_center = 1, do_edges = 0, branch_freq = 20;
                break;
            case 'E':
                all_edges = 1;
            case 'e':
                do_edges = 1;
                break;
            case 'D':
                fade = 1;
                break;
            case 'f':
                season = 1, is_color = 1;
                break;
            case 'H':
                do_halt = 1;
                break;
            case 'r':
                reverse = 1;
                break;
            case 'g':
                dbug = atoi(optarg);
                break;
            case 'G':
                grow_rate = atoi(optarg);
                grow_rate = (grow_rate > 100 ? 100 : grow_rate);
                grow_rate = (grow_rate < 1 ? 1 : grow_rate);
                if (grow_rate != atoi(optarg)) {
                    fprintf(stderr, "Illegal growth rate, using %d instead.\n",
                        grow_rate);
                    fflush(stderr);
                }
                gr_sleep = 1000000 - (10000 * grow_rate);
                break;
#ifndef X11
            case 'w':
                if (is_color)
                    fputs("ignoring -w after -C\n", stderr);
                else
                    do_windows = 1;
#endif /* X11 */
            case 'C':
                if (do_windows)
                    fputs("ignoring -C after -w\n", stderr);
                else
                    is_color = 1;
                break;
            case 'x':
                xsyn = 1;
                break;
            case 'F':
                do_fork = 1;
                break;
            case 'b':
                branch_freq = atoi(optarg);
                branch_freq = (branch_freq > 40 ? 40 : branch_freq);
                branch_freq = (branch_freq < 1 ? 1 : branch_freq);
                if (branch_freq != atoi(optarg)) {
                    fprintf(stderr,
                        "Illegal branch frequency, using %d instead.\n",
                        branch_freq);
                    fflush(stderr);
                }
                break;
            case 'd':
                device = optarg;
                break;
            case 'i':
                interleaf = atoi(optarg);
                break;
            case 'l':
                if (do_edges)
                    fputs("warning: edge vines have constant leaf frequency\n",
                        stderr);
                else {
                    leaf_freq = atoi(optarg);
                    if (leaf_freq > 40)
                        fprintf(stderr, "maximum leaf frequency is 40\n"),
                            errors = 1;
                }
                break;
            case 's':
                lsize = atoi(optarg);   
                if (lsize <= 0)
                    fprintf(stderr, "bad leaf size: %s\n", optarg), errors = 1;
                break;
            case 't':
                technicolor = atoi(optarg);   
                if (technicolor > 256) technicolor = 256;
                if (technicolor < 0) technicolor = 0;
                if (technicolor != atoi(optarg)) {
                    fprintf(stderr,
                        "Illegal number of colors with technicolor, using %d colors instead.\n",
                        technicolor);
                    fflush(stderr);
                }
		is_color = 1;
                break;
            case 'v':
                n_vines = atoi(optarg);
                do_edges = 0;
                break;
            case 'R':
                rot_increment = atoi(optarg);
                rot_increment = (rot_increment > 360 ? 360 : rot_increment);
                rot_increment = (rot_increment < 0 ? 0 : rot_increment);
                if (rot_increment != atoi(optarg)) {
                    fprintf(stderr,
                        "Illegal rotation increment, using %d instead.\n",
                        rot_increment);
                    fflush(stderr);
                }
                break;
            case 'L':
                leaf_file = optarg;
                break;
            case 'S':
                stem_file = optarg;
                break;
            case 'P':
                palette_file = optarg;
                is_color = 1;
                break;
            case '?':
            case 'h':
                errors = 1;
                break;
            case 'z':
                fprintf(stderr, "%s: vines %d.%d (%s %s)\n",
                    prog_name, VERSION, PATCHLEVEL, PATCHDATE, AUTHOR);
                fflush(stderr);
                exit(0);
        }

    if (errors) {
        fprintf(stderr, err_msg[0], prog_name);
        for (c = 1; err_msg[c]; c++)
            fprintf(stderr, "%s\n", err_msg[c]);
        fflush(stderr);
        exit(1);
    }
}



void whine_color(char *file, char *line)
{
    fprintf(stderr, "bad color in file %s: %s", file, line);
    fflush(stderr);
}



#ifdef X11
int get_windows(){ return 0; }



/*
 * Open the display (connection to the X server).  If it's already
 * open, close it and reopen it since this is a child process and
 * each child needs its own connection.
 */
void init_display()
{
    int first_time = !dpy;
    char *getenv(), *p = getenv("DISPLAY");

    if (dpy)
        XCloseDisplay(dpy);
    if (!(dpy = XOpenDisplay(device? device : p? p : ""))) {
        fprintf(stderr, "%s: unable to open display '%s'\n",
            prog_name, XDisplayName(p));
        exit(1);
    }
    if (xsyn)
        XSynchronize(dpy, True);
    XSetErrorHandler(x_error);
    screen = DefaultScreenOfDisplay(dpy);
    win = XRootWindow(dpy,XDefaultScreen(dpy));
    gc = XCreateGC(dpy, win, 0L, NULL);
    init_srandom();
    if (first_time)
        initialize_color();
    XFlush(dpy);
}



int c_depth = 0;
unsigned int c_max = 0;
XColor color;

void initialize_color()
{
    Colormap cm = DefaultColormapOfScreen(screen);
    int real_max_colors = 256;

    if (!is_color)
        return;
    if (c_depth == 0) {
        c_depth = 8;
        c_max = (1 << c_depth) - 1;
    }
    if (technicolor)
        real_max_colors = technicolor;

    color.flags = DoRed | DoGreen | DoBlue;

    if (palette_file) {
        char line[BUFSIZ];
        short r, g, b;

        if (!(f = fopen(palette_file,"r")))
            perror(palette_file), exit(1);
        for (max_color = 0; max_color < real_max_colors; max_color++) {
            if (fgets(line, BUFSIZ, f) == NULL)
                break;
            if (line[0] != '!' && line[0] != '%') {
                if (sscanf(line, "#%2hx%2hx%2hx", &r, &g, &b) < 3) {
                    whine_color(palette_file, line);
                    max_color--;
                } else {
                    color.red = r << c_depth;
                    color.green = g << c_depth;
                    color.blue = b << c_depth;
                    if (dbug)
                        printf("%d=", max_color);
                    if (!XAllocColor(dpy, cm, &color)) {
                        printf("ran out of colors at index %d\n", max_color--);
                        break;
                    }
                    colors[max_color] = color.pixel;
                    if (dbug)
                        printf("%ld, ", color.pixel);
                }
            }
        }
        (void) fclose(f);
        if (max_color > real_max_colors) {
            fprintf(stderr, "%s: too many colors (limit is 256)\n",
                palette_file);
            fflush(stderr);
        } else {
            real_max_colors = max_color - 1;
        }
    } else {
        for (max_color = 0; max_color < real_max_colors; max_color++) {
            if (! technicolor) {
                if (season) {
                    color.green = (c_max - max_color) << c_depth;
                    color.red = (c_max - max_color/3) << c_depth;
                } else {
                    color.red = (c_max - max_color) << c_depth;
                    color.green = (c_max - max_color/3) << c_depth;
                }
                color.blue = (max_color/5) << c_depth;
            } else {
                color.red = (short) ((random() % 256) << c_depth);
                color.green = (short) ((random() % 256) << c_depth);
                color.blue = (short) ((random() % 256) << c_depth);
            }
            if (dbug)
                printf("%d=", max_color);
            if (!XAllocColor(dpy, cm, &color)) {
                fprintf(stderr, "ran out of colors at index %d\n", max_color--);
                fflush(stderr);
                break;
            }
            colors[max_color] = color.pixel;
            if (dbug)
                printf("%ld, ", color.pixel);
        }
    }
}



int x_error(Display *display, XErrorEvent *error)
{
    char        buffer[BUFSIZ];
    char        num[10];
    static int  count;          /* cumulative error total */

    /* get and print major error problem */
    XGetErrorText(display, error->error_code, buffer, BUFSIZ);

    fprintf(stderr, "x_error(%d): <%s>", ++count, buffer);

    /* print the request name that caused the error -- from X Database */
    /* internally this expects a numeric string (I peeked!) */
    sprintf(num, "%d", error->request_code );
    XGetErrorDatabaseText(display, "XRequest", num, num, buffer, BUFSIZ);
    fprintf(stderr, " %s\n", buffer);

    fprintf(stderr, "color.red = %d, green = %d, blue = %d, pixel = %ld\n",
        color.red>>c_depth, color.green>>c_depth, color.blue>>c_depth,
        color.pixel);
    fflush(stderr);
    /* abort to core dump or global debug flag set.. */
    /* abort(0); */
    XCloseDisplay(dpy);
    exit(0);
    return(-1);  /* keep compiler happy */
}



#else /* X11 */

void
init_display()
{
    char groups[PIXPG_OVERLAY+1];
    if (screen)
        return;
    if (!(screen = pr_open(device)))
        exit(1); /* pr_open prints its own error messages (bad idea) */

    pr_available_plane_groups(screen, sizeof(groups), groups);

    if (groups[PIXPG_OVERLAY] && groups[PIXPG_OVERLAY_ENABLE] && !is_color) {
        pr_set_plane_group(screen, PIXPG_OVERLAY_ENABLE);
        pr_rop(screen, 0,0, WidthOfScreen(screen), HeightOfScreen(screen),
            PIX_SET, (Pixrect *) 0,0,0);
        pr_set_plane_group(screen, PIXPG_OVERLAY);
    }
}



int
get_windows()
{
    char name[WIN_NAMESIZE];
    int toolfd, wfd, link;
    struct screen screen;

    /*
     * Determine parent and get root fd
     */
    if (we_getparentwindow(name)) {
        perror(name);
        puts("You need to be running suntools to do the -w option");
        exit(1);
    }
    if ((wfd = open(name, O_RDONLY, 0)) < 0)
        perror(name);
    win_screenget(wfd, &screen);
    close(wfd);
    if ((wfd = open(screen.scr_rootname, O_RDONLY, 0)) < 0)
        perror(screen.scr_rootname);

    link = win_getlink(wfd, WL_OLDESTCHILD);
    while (link != WIN_NULLLINK) {
        win_numbertoname(link, name);
        if ((toolfd = open(name, O_RDONLY, 0)) < 0)
            perror(name), exit(1);
        if (!(win_getuserflags(toolfd) & WMGR_ICONIC))
            win_getrect(toolfd, &win_rects[win_no++]);
        link = win_getlink(toolfd, WL_YOUNGERSIB); /* Next */
        close(toolfd);
    }
    close(wfd);
    if (dbug) {
        int i;
        puts("windows at:");
        for (i = 0; i < win_no; i++) {
            printf("%d,%d size: %d,%d\n",
                win_rects[i].r_left, win_rects[i].r_top,
                win_rects[i].r_width, win_rects[i].r_height);
        }
    }
    return win_no;
}



void
initialize_color()
{
    u_char red[256], green[256], blue[256];

    if (!is_color) {
        cms_monochromeload(red, green, blue);
        pr_putcolormap(screen, 0, 1, red, green, blue);
        max_color = 1;
    } else {
        for (max_color = 0; max_color < 256; max_color++) {
            if (! technicolor) {
                red[max_color] = 255 - max_color;
                green[max_color] = 255 - max_color/3;
                blue[max_color] = max_color/5;
            } else {
                /* caution - untested code! --MEO */
                red[max_color] = (random() % 256);
                green[max_color] = (random() % 256);
                blue[max_color] = (random() % 256);
            }
            colors[max_color] = max_color;
        }
        if (!season)
            pr_putcolormap(screen, 0, 255, red, green, blue);
        else /* swapping red and green maps makes shades of brown */
            pr_putcolormap(screen, 0, 255, green, red, blue);
    }
}


void refresh_root() {}
#endif /* X11 */

